"""Protocol-related functions."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

from emodelrunner.recordings import RecordingCustom
from emodelrunner.locations import SOMA_LOC


logger = logging.getLogger(__name__)

seclist_to_sec = {
    "somatic": "soma",
    "apical": "apic",
    "axonal": "axon",
    "myelinated": "myelin",
}


class CurrentOutputKeyMixin:
    """Contains methods useful for multiple Protocol classes."""

    def curr_output_key(self):
        """Get the output key for current based on the one for voltage.

        Returns:
            str used as key in current dict
        """
        if self.recordings is not None:
            # this gives 'prefix.name'
            name = ".".join(self.recordings[0].name.split(".")[:2])
        else:
            name = ""
        return "current_" + name


def get_extra_recording_location(recording_definition, apical_point_isec=-1):
    """Get the location for the extra recording.

    Args:
        recording_definition (dict): contains the extra recording configuration data
        apical_point_isec (int): apical point section index.
            Should be given if the recording definition "type" is "somadistanceapic"

    Raises:
        Exception: if the recording definition "type" is "somadistanceapic" and
            apical_point_isec is -1.
        Exception: if the 'type' in the recording definition is neither
            "somadistance", nor "somadistanceapic", nor "nrnseclistcomp"

    Returns:
        location of the extra recording
    """
    if recording_definition["type"] == "somadistance":
        location = ephys.locations.NrnSomaDistanceCompLocation(
            name=recording_definition["name"],
            soma_distance=recording_definition["somadistance"],
            seclist_name=recording_definition["seclist_name"],
        )

    elif recording_definition["type"] == "somadistanceapic":
        if apical_point_isec == -1:
            raise Exception(
                "Cannot record at a given distance from apical point"
                f"if apical_point_isec is {apical_point_isec}."
            )
        location = ephys.locations.NrnSecSomaDistanceCompLocation(
            name=recording_definition["name"],
            soma_distance=recording_definition["somadistance"],
            seclist_name=seclist_to_sec[recording_definition["seclist_name"]],
            sec_index=apical_point_isec,
        )

    elif recording_definition["type"] == "nrnseclistcomp":
        location = ephys.locations.NrnSeclistCompLocation(
            name=recording_definition["name"],
            comp_x=recording_definition["comp_x"],
            sec_index=recording_definition["sec_index"],
            seclist_name=recording_definition["seclist_name"],
        )

    else:
        raise Exception(f"Recording type {recording_definition['type']} not supported")

    return location


def get_recordings(protocol_name, protocol_definition, prefix, apical_point_isec=-1):
    """Get recordings from protocol definition.

    Args:
        protocol_name (str): name of the protocol
        protocol_definition (dict): dict containing the protocol data
        prefix (str): prefix used in naming responses, features, recordings, etc.
        apical_point_isec (int): apical point section index
            Should be given if there is "somadistanceapic" in "type"
            of at least one of the extra recording definition

    Returns:
        list of RecordingCustom
    """
    recordings = []
    recordings.append(
        RecordingCustom(
            name=f"{prefix}.{protocol_name}.soma.v",
            location=SOMA_LOC,
            variable="v",
        )
    )

    if "extra_recordings" in protocol_definition:
        for recording_definition in protocol_definition["extra_recordings"]:

            location = get_extra_recording_location(
                recording_definition, apical_point_isec
            )

            var = recording_definition["var"]
            recording = RecordingCustom(
                name=f"{prefix}.{protocol_name}.{location.name}.{var}",
                location=location,
                variable=var,
            )
            recordings.append(recording)

    return recordings


def check_for_forbidden_protocol(protocols_dict):
    """Check for unsupported protocol.

    Args:
        protocols_dict (dict): contains all protocols to be run

    Raises:
        Exception: If a protocol that should only be used with MainProtocol is present
            in protocols_dict
    """
    # Those protocols cannot be used if they are not in MainProtocol
    forbidden_prots = [
        "RatSSCxRinHoldcurrentProtocol",
        "RatSSCxThresholdDetectionProtocol",
        "StepThresholdProtocol",
        "RampThresholdProtocol",
    ]
    # check the class name of each protocol
    for prot in protocols_dict.values():
        if type(prot).__name__ in forbidden_prots:
            prot_name = type(prot).__name__
            raise Exception(
                "No MainProtocol found, but {prot} was found."
                f"To use {prot_name}, please set MainProtocol."
            )



"""Computation of morphology features."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np
import neurom as nm
from neurom.core.morphology import iter_neurites
from neurom.core.types import tree_type_checker as is_type
from neurom.features.neurite import segment_radii, segment_lengths

logger = logging.getLogger(__name__)


class MorphologyFeature:
    """Morphology feature representation.

    Attributes:
        name (str): name of the feature
        value (float or list): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self):
        """Sets initial attribute values."""
        self.name = None
        self.value = None
        self.unit = None

    @staticmethod
    def replace_empty_value(value):
        """Replaces the empty value with 0 or [0].

        Attrs:
            value (float or list): the value to be replaced if empty

        Returns:
            float/int or non-empty list
        """
        if value is None:
            value = 0
        elif hasattr(value, "len") and len(value) == 0:
            value = [0]
        return value

    def to_dict(self):
        """Returns a dictionary from the fields."""
        return self.__dict__


def _seg_lengths(neurite):
    """Syntactic sugar to retrieve segment_lengths as np.array."""
    return np.array(segment_lengths(neurite))


def _seg_radii(neurite):
    """Syntactic sugar to retrieve segment_radii as np.array."""
    return np.array(segment_radii(neurite))


class AverageDiameter(MorphologyFeature):
    """Average diameter (weighted by section length) feature for a neurite type."""

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"average diameter of {neurite_name}"
        self.unit = "\u00b5m"

        neurites = list(iter_neurites(morphology, filt=is_type(neurite_type)))

        neurite_weighted_radii = [_seg_radii(n) * _seg_lengths(n) for n in neurites]
        neurite_seg_lengths = [_seg_lengths(n) for n in neurites]

        neurite_weighted_radii = np.concatenate(neurite_weighted_radii, axis=0)
        neurite_seg_lengths = np.concatenate(neurite_seg_lengths, axis=0)

        avg_radius = np.sum(neurite_weighted_radii) / np.sum(neurite_seg_lengths)
        avg_diameter = avg_radius * 2
        self.value = avg_diameter


class TotalLength(MorphologyFeature):
    """Total length feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} length"
        self.unit = "\u00b5m"
        feature_value = nm.get("total_length", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class TotalHeight(MorphologyFeature):
    """Total height feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} height"
        self.unit = "\u00b5m"
        feature_value = nm.get("total_height", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class TotalWidth(MorphologyFeature):
    """Total width feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} width"
        self.unit = "\u00b5m"
        feature_value = nm.get("total_width", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class TotalDepth(MorphologyFeature):
    """Total depth feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} depth"
        self.unit = "\u00b5m"
        feature_value = nm.get("total_depth", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class TotalArea(MorphologyFeature):
    """Total area feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} area"
        self.unit = "\u00b5m\u00b2"
        feature_value = nm.get("total_area", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class TotalVolume(MorphologyFeature):
    """Total volume feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"total {neurite_name} volume"
        self.unit = "\u00b5m\u00b3"
        feature_value = nm.get("total_volume", morphology, neurite_type=neurite_type)
        self.value = self.replace_empty_value(feature_value)


class NumberOfSections(MorphologyFeature):
    """Number of sections feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"number of {neurite_name} sections"
        self.unit = ""
        feature_value = nm.get(
            "number_of_sections", morphology, neurite_type=neurite_type
        )
        self.value = self.replace_empty_value(feature_value)


class NumberOfSegments(MorphologyFeature):
    """Number of segments feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"number of {neurite_name} segments"
        self.unit = ""
        feature_values = nm.get(
            "number_of_segments", morphology, neurite_type=neurite_type
        )
        self.value = self.replace_empty_value(feature_values)


class MeanNeuriteVolumes(MorphologyFeature):
    """Mean neurite volume feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"mean {neurite_name} volume"
        self.unit = "\u00b5m\u00b3"
        feature_values = nm.get(
            "total_volume_per_neurite", morphology, neurite_type=neurite_type
        )
        feature_values = self.replace_empty_value(feature_values)
        self.value = sum(feature_values) / len(feature_values)


class AverageSectionLength(MorphologyFeature):
    """Average section length feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"average {neurite_name} section length"
        self.unit = "\u00b5m"
        feature_values = nm.get(
            "section_lengths", morphology, neurite_type=neurite_type
        )
        feature_values = self.replace_empty_value(feature_values)
        self.value = sum(feature_values) / len(feature_values)


class AverageSegmentLength(MorphologyFeature):
    """Average segment length feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"average {neurite_name} segment length"
        self.unit = "\u00b5m"
        feature_values = nm.get(
            "segment_lengths", morphology, neurite_type=neurite_type
        )
        feature_values = self.replace_empty_value(feature_values)
        self.value = sum(feature_values) / len(feature_values)


class MaxBranchOrder(MorphologyFeature):
    """Maximum branch order feature.

    Attributes:
        name (str): name of the feature
        value (int): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"{neurite_name} maximum branch order"
        self.unit = ""
        feature_values = nm.get(
            "section_branch_orders", morphology, neurite_type=neurite_type
        )
        feature_values = self.replace_empty_value(feature_values)
        self.value = max(feature_values)


class MaxSectionLength(MorphologyFeature):
    """Maximum section length feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology, neurite_name, neurite_type):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
            neurite_name (str): neurite name, e.g. axon
            neurite_type (NeuriteType): enum for neurite type encoding
        """
        super().__init__()
        self.name = f"{neurite_name} maximum section length"
        self.unit = "\u00b5m"
        feature_values = nm.get(
            "section_lengths", morphology, neurite_type=neurite_type
        )
        feature_values = self.replace_empty_value(feature_values)
        self.value = max(feature_values)


class SomaDiamater(MorphologyFeature):
    """Soma diameter feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
        """
        super().__init__()
        self.name = "soma diameter"
        self.unit = "\u00b5m"
        feature_value = nm.get("soma_radius", morphology)
        self.value = 2 * self.replace_empty_value(feature_value)


class SomaSurfaceArea(MorphologyFeature):
    """Soma surface area feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
        """
        super().__init__()
        self.name = "soma surface area"
        self.unit = "\u00b5m\u00b2"
        feature_value = nm.get("soma_surface_area", morphology)
        self.value = self.replace_empty_value(feature_value)


class SomaVolume(MorphologyFeature):
    """Soma volume feature.

    Attributes:
        name (str): name of the feature
        value (float): value of the feature
        unit (str): unit of the feature
    """

    def __init__(self, morphology):
        """Constructor.

        Args:
            morphology (neurom neuron object): morphology object
        """
        super().__init__()
        self.name = "soma volume"
        self.unit = "\u00b5m\u00b3"
        feature_value = nm.get("soma_volume", morphology)
        self.value = self.replace_empty_value(feature_value)


class MorphologyFactsheetBuilder:
    """Computes the factsheet values for a morphology.

    Attributes:
        morphology (neurom.core.morphology.Morphology): morphology of the neuron
        neurites (list): list of neurites to be considered
        neurite_features (list): list of neurite feature to be used
        soma_features (list): list of soma features to be used
    """

    def __init__(self, morph_path):
        """Load the morphology.

        Args:
            morph_path (str or Path): Path to the morphology file.
        """
        self.morphology = nm.load_morphology(morph_path)
        self.neurites = []
        self.neurite_features = []
        self.soma_features = []

    def get_neurites(self):
        """Return neurite names (str) and types (neurom type).

        If basal or apical are not present, name them 'dendrite'.


        Returns:
            list of tuple: (neurite_names, neurite_types)
        """
        api = nm.get("total_length", self.morphology, neurite_type=nm.APICAL_DENDRITE)
        bas = nm.get("total_length", self.morphology, neurite_type=nm.BASAL_DENDRITE)
        if api and bas:
            return [
                ("axon", nm.AXON),
                ("apical", nm.APICAL_DENDRITE),
                ("basal", nm.BASAL_DENDRITE),
            ]
        elif api:
            return [("axon", nm.AXON), ("dendrite", nm.APICAL_DENDRITE)]
        elif bas:
            return [("axon", nm.AXON), ("dendrite", nm.BASAL_DENDRITE)]
        logger.warning("No dendrite found!")
        return [("axon", nm.AXON)]

    def get_feature_values(self):
        """Returns the values of all features in a list."""
        neurites = self.neurites
        all_values = []
        for neurite_name, neurite_type in neurites:
            for feature in self.neurite_features:
                feature_dict = feature(
                    self.morphology, neurite_name, neurite_type
                ).to_dict()
                all_values.append(feature_dict)

        for feature in self.soma_features:
            feature_dict = feature(self.morphology).to_dict()
            all_values.append(feature_dict)

        return all_values

    def factsheet_dict(self):
        """Returns the factsheet as a dict."""
        anatomy = self.get_feature_values()
        return {"name": "Anatomy", "values": anatomy}


class SSCXMorphologyFactsheetBuilder(MorphologyFactsheetBuilder):
    """Morphology factsheet builder for SSCX packages."""

    def __init__(self, morph_path):
        """Load the morphology.

        Args:
            morph_path (str or Path): Path to the morphology file.
        """
        super().__init__(morph_path)
        self.neurites = self.get_neurites()
        self.neurite_features = [
            TotalLength,
            MeanNeuriteVolumes,
            MaxBranchOrder,
            MaxSectionLength,
        ]
        self.soma_features = [SomaDiamater]


class HippocampusMorphologyFactsheetBuilder(MorphologyFactsheetBuilder):
    """Morphology factsheet builder for Hippocampus packages."""

    def __init__(self, morph_path):
        """Load the morphology.

        Args:
            morph_path (str or Path): Path to the morphology file.
        """
        super().__init__(morph_path)
        self.neurites = [("all", nm.ANY_NEURITE)] + self.get_neurites()
        self.neurite_features = [
            TotalWidth,
            TotalHeight,
            TotalDepth,
            TotalLength,
            TotalArea,
            TotalVolume,
            AverageDiameter,
            NumberOfSections,
            NumberOfSegments,
            AverageSectionLength,
            AverageSegmentLength,
            MaxBranchOrder,
        ]
        self.soma_features = [SomaDiamater, SomaSurfaceArea, SomaVolume]


class ThalamusMorphologyFactsheetBuilder(MorphologyFactsheetBuilder):
    """Morphology factsheet builder for Thalamus packages."""

    def __init__(self, morph_path):
        """Load the morphology.

        Args:
            morph_path (str or Path): Path to the morphology file.
        """
        super().__init__(morph_path)
        self.neurites = self.get_neurites()
        self.neurite_features = [
            TotalLength,
            TotalVolume,
            MaxBranchOrder,
        ]
        self.soma_features = [SomaDiamater, SomaSurfaceArea, SomaVolume]



"""Configuration parsing."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from configparser import ConfigParser
from enum import Enum

from emodelrunner.configuration.subgroups import (
    HocPaths,
    ProtArgs,
    SynMechArgs,
    MorphArgs,
    PresynStimArgs,
)


class PackageType(Enum):
    """Enumerator for the emodel package types."""

    sscx = "sscx"
    thalamus = "thalamus"
    synplas = "synplas"


class EModelConfigParser(ConfigParser):
    """Built-in ConfigParser annotated with package type."""

    def __init__(self):
        """Constructor."""
        super().__init__()

    @property
    def package_type(self):
        """Package type as a property."""
        return PackageType[self.get("Package", "type")]

    def hoc_paths_args(self):
        """Get the config data subgroup containing the paths to the hoc files."""
        return HocPaths(
            hoc_dir=self.get("Paths", "memodel_dir"),
            cell_hoc_filename=self.get("Paths", "cell_hoc_file"),
            simul_hoc_filename=self.get("Paths", "simul_hoc_file"),
            run_hoc_filename=self.get("Paths", "run_hoc_file"),
            syn_dir=self.get("Paths", "syn_dir"),
            syn_dir_for_hoc=self.get("Paths", "syn_dir_for_hoc"),
            syn_hoc_filename=self.get("Paths", "syn_hoc_file"),
            main_protocol_filename=self.get("Paths", "main_protocol_file"),
        )

    def prot_args(self):
        """Get the subgroup containing protocols configuration data."""
        return ProtArgs(
            emodel=self.get("Cell", "emodel"),
            apical_point_isec=self.getint("Protocol", "apical_point_isec"),
            mtype=self.get("Morphology", "mtype"),
            prot_path=self.get("Paths", "prot_path"),
            features_path=self.get("Paths", "features_path"),
        )

    def syn_mech_args(self, add_synapses=None, seed=None, rng_settings_mode=None):
        """Get the data from config used when loading synapse mechanisms."""
        if add_synapses is None:
            add_synapses = self.getboolean("Synapses", "add_synapses")
        if seed is None:
            seed = self.getint("Synapses", "seed")
        if rng_settings_mode is None:
            rng_settings_mode = self.get("Synapses", "rng_settings_mode")

        return SynMechArgs(
            add_synapses=add_synapses,
            seed=seed,
            rng_settings_mode=rng_settings_mode,
            syn_conf_file=self.get("Paths", "syn_conf_file"),
            syn_data_file=self.get("Paths", "syn_data_file"),
            syn_dir=self.get("Paths", "syn_dir"),
        )

    def morph_args(self):
        """Get morphology arguments for SSCX from the configuration object."""
        morph_args = {}
        morph_args["morph_path"] = self.get("Paths", "morph_path")
        morph_args["do_replace_axon"] = self.getboolean("Morphology", "do_replace_axon")

        if self.package_type == PackageType.sscx:
            morph_args["axon_hoc_path"] = self.get("Paths", "replace_axon_hoc_path")

        return MorphArgs(**morph_args)

    def synplas_morph_args(self, precell=False):
        """Get morphology arguments for Synplas from the configuration object.

        Args:
            precell (bool): True to load precell morph. False to load usual morph.
        """
        # load morphology path
        if precell:
            morph_path = self.get("Paths", "precell_morph_path")
        else:
            morph_path = self.get("Paths", "morph_path")

        morph_args = {
            "morph_path": morph_path,
            "do_replace_axon": self.getboolean("Morphology", "do_replace_axon"),
        }
        return MorphArgs(**morph_args)

    def presyn_stim_args(self, pre_spike_train):
        """Get the pre-synaptic stimulus config data.

        Args:
            pre_spike_train (list): times at which the synapses fire (ms)
        """
        # spikedelay is the time between the start of the stimulus
        # and the precell spike time
        spike_delay = self.getfloat("Protocol", "precell_spikedelay")

        # stim train is the times at which to stimulate the precell
        return PresynStimArgs(
            stim_train=pre_spike_train - spike_delay,
            amp=self.getfloat("Protocol", "precell_amplitude"),
            width=self.getfloat("Protocol", "precell_width"),
        )



"""GUI."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from emodelrunner.GUI_utils.interface import GUI
from emodelrunner.parsing_utilities import get_parser_args, set_verbosity

logger = logging.getLogger(__name__)

if __name__ == "__main__":
    args = get_parser_args()
    set_verbosity(args.verbosity)

    gui = GUI(config_path=args.config_path)
    gui.root.mainloop()



"""Builds morphology objects."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from emodelrunner.morphology.morphology import (
    SSCXNrnFileMorphology,
    ThalamusNrnFileMorphology,
)
from emodelrunner.configuration import PackageType


def get_axon_hoc(axon_hoc_path):
    """Returns string containing axon hoc to use as replacement.

    Args:
        axon_hoc_path (str): path to axon hoc file

    Returns:
        str: new axon hoc
    """
    with open(axon_hoc_path, "r", encoding="utf-8") as f:
        return f.read()


def create_morphology(morph_args, package_type):
    """Creates the morphology object.

    Args:
        morph_args (MorphArgs): morphology-related configuration
        package_type (Enum): enum denoting the package type

    Raises:
        ValueError: raised when package_type is not supported

    Returns:
        ephys.morphologies.NrnFileMorphology: morphology object
    """
    replace_axon_hoc = None
    if morph_args.axon_hoc_path is not None:
        replace_axon_hoc = get_axon_hoc(morph_args.axon_hoc_path)

    if package_type in [PackageType.sscx, PackageType.synplas]:
        morph = SSCXNrnFileMorphology(
            morph_args.morph_path,
            do_replace_axon=morph_args.do_replace_axon,
            replace_axon_hoc=replace_axon_hoc,
        )
    elif package_type == PackageType.thalamus:
        morph = ThalamusNrnFileMorphology(
            morph_args.morph_path,
            do_replace_axon=morph_args.do_replace_axon,
            replace_axon_hoc=replace_axon_hoc,
        )
    else:
        raise ValueError(f"unsupported package type: {package_type}")
    return morph



"""To extract ion channel mechanisms information from static parameter files."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re


logger = logging.getLogger(__name__)


def edit_dist_func(value):
    """Edit function expression to be latex and plot readable.

    Args:
        value (str): the distribution function expressed as a string.
            ex: "(-0.8696 + 2.087*math.exp((x)*0.0031))*3.1236056887012746e-06"

    Returns:
        a tuple containing

        - latex (str): a latex-compatible version of the distrib. function
          ex: "(-0.8696 + 2.087*e^{x*0.0031})*3.1236056887012746e-06"
        - value (str): a plottable version of the distrib. function
          ex: "(-0.8696 + 2.087exp(x*0.0031))*3.1236056887012746e-06"
    """
    if "math" in value:
        value = value.replace("math.", "")
    if "(x)" in value:
        value = value.replace("(x)", "x")
    latex = re.sub(r"exp*\(([0-9x*-.]*)\)", "e^{\\1}", value)
    return latex, value


def get_channel_and_equations(
    name, param_config, full_name, exp_fun, decay_fun, release_params
):
    """Returns the channel and a dictionary containing equation type (uniform or exp) and value.

    Args:
        name (str): the name of the ion channel and its parameter.
            should have the form parameter_channel (ex: "gCa_HVAbar_Ca_HVA2")
        param_config (dict): parameter dictionary taken from parameter data file.
            should have a "dist" key if the distribution is exponential.
        full_name (str): should have the form name.section
        exp_fun (str): the distribution function expressed as a string.
            ex: "(-0.8696 + 2.087*math.exp((x)*0.0031))*3.1236056887012746e-06"
        decay_fun (str): the decay function expressed as a string.
        release_params (dict): final parameters of the optimised cell.

    Returns:
        a tuple containing

        - channel (str): name of the channel (ex: "Ca_HVA2")
        - biophys (str): parameter name (ex: "gCa_HVAbar")
        - equation_dict (dict): contains equation values and equation type (uniform or exp)
    """
    # parameter value (obtained from optimisation)
    value = release_params[full_name]

    decay_cst = None
    if decay_fun:
        decay_cst = release_params["constant.distribution_decay"]

    # isolate channel and biophys
    split_name = name.split("_")
    if len(split_name) == 4:
        biophys = "_".join(split_name[0:2])
        channel = "_".join(split_name[2:4])
    elif len(split_name) == 3:
        biophys = split_name[0]
        channel = "_".join(split_name[1:3])
    elif len(split_name) == 2:
        biophys = split_name[0]
        channel = split_name[1]

    # type
    if "dist" in param_config:
        if param_config["dist"] == "exp":
            type_ = "exponential"
            value = exp_fun.format(distance="x", value=value)
            latex, plot = edit_dist_func(value)
        elif param_config["dist"] == "decay":
            type_ = "decay"
            value = decay_fun.format(distance="x", value=value, constant=decay_cst)
            latex, plot = edit_dist_func(value)
        else:
            logger.warning(
                "dist is set to %s. Expected 'exp' or 'decay'. Set type to exponential anyway.",
                param_config["dist"],
            )
    else:
        type_ = "uniform"
        latex = value
        plot = value

    return channel, biophys, {"latex": latex, "plot": plot, "type": type_}


def append_equation(location_map, section, channel, biophys, equation_dict):
    """Append equation to location map dict.

    Args:
        location_map (dict): contains the equations for each channel at each location.
            To be filled by this function
        section (str): the section(s) to which the equations should be appended
            Can be
            "alldend", "somadend", "somaxon", "allact",
            "apical", "basal", "axonal" or "somatic"
        channel (str): name of the channel (ex: "Ca_HVA2")
        biophys (str): parameter name (ex: "gCa_HVAbar")
        equation_dict (dict): dictionary containing equation values (for plotting and latex display)
            and type (uniform or exponential)
    """
    # do not take into account "all" section
    # set default to create keys then add equations
    # this allows to either create channel data or append to channel key
    if section == "alldend":
        location_map["all dendrites"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["all dendrites"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "somadend":
        location_map["all dendrites"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["all dendrites"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
        location_map["somatic"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["somatic"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "somaxon":
        location_map["somatic"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["somatic"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
        location_map["axonal"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["axonal"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "allact":
        location_map["all dendrites"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["all dendrites"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
        location_map["somatic"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["somatic"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
        location_map["axonal"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["axonal"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "apical":
        location_map["apical"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["apical"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "basal":
        location_map["basal"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["basal"]["channels"][channel]["equations"][biophys] = equation_dict
    elif section == "axonal":
        location_map["axonal"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["axonal"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict
    elif section == "somatic":
        location_map["somatic"]["channels"].setdefault(
            channel, {"equations": {biophys: equation_dict}}
        )
        location_map["somatic"]["channels"][channel]["equations"][
            biophys
        ] = equation_dict


def clean_location_map(location_map):
    """Remove empty locations.

    Args:
        location_map (dict): contains the equations for each channel at each location
    """
    for key, loc in list(location_map.items()):
        if not loc["channels"]:
            location_map.pop(key)


def get_mechanisms_data(emodel, optimized_params_dict, unoptimized_params_dict):
    """Return a dictionary containing channel mechanisms for each section.

    Args:
        emodel (str): name of the emodel
        optimized_params_dict (dict): contains the optimized parameters
        unoptimized_params_dict (dict): contains the unoptimized parameters,
            and also contains the decay and exponential equations

    Returns:
        dict: containing the channel mechanisms
    """
    # pylint: disable=too-many-locals
    release_params = optimized_params_dict[emodel]["params"]

    decay_func = None
    if "decay" in unoptimized_params_dict["distributions"].keys():
        decay_func = unoptimized_params_dict["distributions"]["decay"]["fun"]

    parameters = unoptimized_params_dict["parameters"]
    exp_fun = unoptimized_params_dict["distributions"]["exp"]["fun"]

    location_map = {
        "all dendrites": {"channels": {}},
        "apical": {"channels": {}},
        "basal": {"channels": {}},
        "somatic": {"channels": {}},
        "axonal": {"channels": {}},
    }

    for section, unoptimized_params_list in parameters.items():
        # do not take into account "comment"
        if isinstance(unoptimized_params_list, list):
            for param_config in unoptimized_params_list:
                name = param_config["name"]
                full_name = ".".join((name, section))

                # only take into account parameters present in finals.json
                if (
                    full_name in release_params
                    and full_name != "constant.distribution_decay"
                ):
                    channel, biophys, equation_dict = get_channel_and_equations(
                        name,
                        param_config,
                        full_name,
                        exp_fun,
                        decay_func,
                        release_params,
                    )

                    # append equation in location map
                    append_equation(
                        location_map, section, channel, biophys, equation_dict
                    )

    # remove empty locations
    clean_location_map(location_map)

    values = [
        {
            "tooltip": "",
            "location_map": location_map,
            "unit": "",
            "name": "list of ion channel mechanisms",
        }
    ]
    return {"name": "Channel mechanisms", "values": values}



"""Create python recordings."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from bluepyopt import ephys

from emodelrunner.configuration.configparser import PackageType
from emodelrunner.create_cells import create_cell_using_config
from emodelrunner.parsing_utilities import get_parser_args, set_verbosity
from emodelrunner.protocols.create_protocols import ProtocolBuilder
from emodelrunner.load import (
    load_config,
    get_release_params,
)
from emodelrunner.output import write_current
from emodelrunner.output import write_responses

logger = logging.getLogger(__name__)


def main(config_path):
    """Main.

    Args:
        config_path (str): path to config file
            The config file should have '.ini' suffix
    """
    # pylint: disable=too-many-locals
    config = load_config(config_path=config_path)

    cell = create_cell_using_config(config)
    release_params = get_release_params(config)

    cvode_active = config.getboolean("Sim", "cvode_active")

    # simulator
    dt = config.getfloat("Sim", "dt")
    sim = ephys.simulators.NrnSimulator(dt=dt, cvode_active=cvode_active)

    # create protocols
    add_synapses = config.getboolean("Synapses", "add_synapses")
    prot_args = config.prot_args()

    if config.package_type == PackageType.sscx:
        protocols = ProtocolBuilder.using_sscx_protocols(add_synapses, prot_args, cell)
    elif config.package_type == PackageType.thalamus:
        protocols = ProtocolBuilder.using_thalamus_protocols(
            add_synapses, prot_args, cell
        )
    else:
        raise ValueError(f"unsupported package type: {config.package_type}")
    ephys_protocols = protocols.get_ephys_protocols()

    # run
    logger.info("Python Recordings Running...")
    responses = ephys_protocols.run(
        cell_model=cell, param_values=release_params, sim=sim, isolate=False
    )

    mtype = config.get("Morphology", "mtype")
    if config.package_type == PackageType.sscx:
        currents = protocols.get_stim_currents(responses, dt)
    elif config.package_type == PackageType.thalamus:
        currents = protocols.get_thalamus_stim_currents(responses, mtype, dt)

    # write responses
    output_dir = config.get("Paths", "output_dir")
    write_responses(responses, output_dir)
    write_current(currents, output_dir)

    logger.info("Python Recordings Done")


if __name__ == "__main__":
    args = get_parser_args()
    set_verbosity(args.verbosity)

    main(config_path=args.config_path)



"""Functions mainly for loading params for run.py."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import json

from bluepyopt import ephys

from emodelrunner.synapses.mechanism import NrnMODPointProcessMechanismCustom
from emodelrunner.locations import multi_locations
from emodelrunner.configuration import get_validated_config


def load_config(config_path):
    """Returns the validated configuration file.

    Args:
        config_path (str or Path): path to the configuration file.

    Returns:
        configparser.ConfigParser: loaded config object
    """
    return get_validated_config(config_path)


def load_emodel_params(emodel, params_path):
    """Get optimized parameters.

    Args:
        emodel (str): name of the emodel
        params_path (str): path to the optimized parameters json file

    Returns:
        dict: optimized parameters for the given emodel
    """
    with open(params_path, "r", encoding="utf-8") as params_file:
        params = json.load(params_file)

    param_dict = params[emodel]["params"]

    return param_dict


def get_syn_setup_params(
    syn_extra_params_path,
    cpre_cpost_path,
    fit_params_path,
    gid,
    invivo,
):
    """Load json files and return syn_setup_params dict.

    Args:
        syn_extra_params_path (str): path to the glusynapses related extra parameters file
        cpre_cpost_path (str): path to the c_pre and c_post related file
            c_pre (resp. c_post) is the calcium amplitude during isolated presynaptic
            (resp. postsynaptic) activation
        fit_params_path (str): path to the file containing the glusynapse fitted parameters
            The fitted parameters are time constant of calcium integrator, depression rate,
            potentiation rate, and factors used in plasticity threshold computation.
        gid (int): ID of the postsynaptic cell
        invivo (bool): whether to run the simulation in 'in vivo' conditions

    Returns:
        dict: glusynapse setup related parameters
    """
    with open(syn_extra_params_path, "r", encoding="utf-8") as f:
        syn_extra_params = json.load(f)
    with open(cpre_cpost_path, "r", encoding="utf-8") as f:
        cpre_cpost = json.load(f)
    with open(fit_params_path, "r", encoding="utf-8") as f:
        fit_params = json.load(f)

    return {
        "syn_extra_params": syn_extra_params,
        "c_pre": cpre_cpost["c_pre"],
        "c_post": cpre_cpost["c_post"],
        "fit_params": fit_params,
        "postgid": gid,
        "invivo": invivo,
    }


def get_release_params(config, precell=False):
    """Return the final parameters.

    Args:
        config (configparser.ConfigParser): configuration
        precell (bool): True to load precell optimized parameters. False to get usual parameters.

    Returns:
        dict: optimized parameters
    """
    if precell:
        emodel = config.get("Cell", "precell_emodel")
    else:
        emodel = config.get("Cell", "emodel")
    params_path = config.get("Paths", "params_path")

    release_params = load_emodel_params(params_path=params_path, emodel=emodel)

    return release_params


def load_mechanisms(mechs_path):
    """Define mechanisms.

    Args:
        mechs_path (str): path to the unoptimized parameters json file

    Returns:
        list of ephys.mechanisms.NrnMODMechanism from file
    """
    with open(mechs_path, "r", encoding="utf-8") as mechs_file:
        mechs = json.load(mechs_file)
    mech_definitions = mechs["mechanisms"]

    mechanisms_list = []
    for sectionlist, channels in mech_definitions.items():

        seclist_locs = multi_locations(sectionlist)

        for channel in channels["mech"]:
            mechanisms_list.append(
                ephys.mechanisms.NrnMODMechanism(
                    name=f"{channel}.{sectionlist}",
                    mod_path=None,
                    suffix=channel,
                    locations=seclist_locs,
                    preloaded=True,
                )
            )

    return mechanisms_list


def load_unoptimized_parameters(params_path, v_init, celsius):
    """Load unoptimized parameters as BluePyOpt parameters.

    Args:
        params_path (str): path to the json file containing
            the non-optimised parameters
        v_init (int): initial voltage (mV). Will override v_init value from parameter file
        celsius (int): cell temperature in celsius.
            Will override celsius value from parameter file

    Returns:
        list of parameters
    """
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    parameters = []

    with open(params_path, "r", encoding="utf-8") as params_file:
        definitions = json.load(params_file)

    # set distributions
    distributions = collections.OrderedDict()
    distributions["uniform"] = ephys.parameterscalers.NrnSegmentLinearScaler()

    distributions_definitions = definitions["distributions"]
    for distribution, definition in distributions_definitions.items():

        if "parameters" in definition:
            dist_param_names = definition["parameters"]
        else:
            dist_param_names = None
        distributions[
            distribution
        ] = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
            name=distribution,
            distribution=definition["fun"],
            dist_param_names=dist_param_names,
        )

    params_definitions = definitions["parameters"]

    if "__comment" in params_definitions:
        del params_definitions["__comment"]

    for sectionlist, params in params_definitions.items():
        if sectionlist == "global":
            seclist_locs = None
            is_global = True
            is_dist = False
        elif "distribution_" in sectionlist:
            is_dist = True
            seclist_locs = None
            is_global = False
            dist_name = sectionlist.split("distribution_")[1]
            dist = distributions[dist_name]
        else:
            seclist_locs = multi_locations(sectionlist)
            is_global = False
            is_dist = False

        bounds = None
        value = None
        for param_config in params:
            param_name = param_config["name"]

            if isinstance(param_config["val"], (list, tuple)):
                is_frozen = False
                bounds = param_config["val"]
                value = None

            else:
                is_frozen = True
                value = param_config["val"]
                bounds = None

            if is_global:
                # force v_init to the given value
                if param_name == "v_init":
                    value = v_init
                elif param_name == "celsius":
                    value = celsius
                parameters.append(
                    ephys.parameters.NrnGlobalParameter(
                        name=param_name,
                        param_name=param_name,
                        frozen=is_frozen,
                        bounds=bounds,
                        value=value,
                    )
                )
            elif is_dist:
                parameters.append(
                    ephys.parameters.MetaParameter(
                        name=f"{param_name}.{sectionlist}",
                        obj=dist,
                        attr_name=param_name,
                        frozen=is_frozen,
                        bounds=bounds,
                        value=value,
                    )
                )

            else:
                if "dist" in param_config:
                    dist = distributions[param_config["dist"]]
                    use_range = True
                else:
                    dist = distributions["uniform"]
                    use_range = False

                if use_range:
                    parameters.append(
                        ephys.parameters.NrnRangeParameter(
                            name=f"{param_name}.{sectionlist}",
                            param_name=param_name,
                            value_scaler=dist,
                            value=value,
                            bounds=bounds,
                            frozen=is_frozen,
                            locations=seclist_locs,
                        )
                    )
                else:
                    parameters.append(
                        ephys.parameters.NrnSectionParameter(
                            name=f"{param_name}.{sectionlist}",
                            param_name=param_name,
                            value_scaler=dist,
                            value=value,
                            bounds=bounds,
                            frozen=is_frozen,
                            locations=seclist_locs,
                        )
                    )

    return parameters


def get_rin_exp_voltage_base(features_path):
    """Get experimental rin voltage base from feature file when having MainProtocol."""
    with open(features_path, "r", encoding="utf-8") as features_file:
        feature_definitions = json.load(features_file)

    rin_exp_voltage_base = None
    for feature in feature_definitions["Rin"]["soma.v"]:
        if feature["feature"] == "voltage_base":
            rin_exp_voltage_base = feature["val"][0]

    if rin_exp_voltage_base is None:
        raise KeyError(f"No voltage_base feature found for 'Rin' in {features_path}")

    return rin_exp_voltage_base


def load_syn_mechs(
    seed,
    rng_settings_mode,
    syn_data_path,
    syn_conf_path,
    pre_mtypes=None,
    stim_params=None,
    use_glu_synapse=False,
    syn_setup_params=None,
):
    """Load synapse mechanisms.

    Args:
        seed (int): random number generator seed number
        rng_settings_mode (str): mode of the random number generator
            Can be "Random123" or "Compatibility"
        syn_data_path (str): path to the (tsv) synapses data file
        syn_conf_path (str): path to the synapse configuration data file
        pre_mtypes (list of ints): activate only synapses whose pre_mtype
            is in this list. if None, all synapses are activated
        stim_params (dict or None): dict with pre_mtype as key,
            and netstim params list as item.
            netstim params list is [start, interval, number, noise]
        use_glu_synapse (bool): if True, instantiate synapses to use GluSynapse
        syn_setup_params (dict): contains extra parameters to setup synapses
            when using GluSynapseCustom

    Returns:
        NrnMODPointProcessMechanismCustom: the synapses mechanisms
    """
    # load synapse file data
    synapses_data = load_synapses_tsv_data(syn_data_path)

    # load synapse configuration
    synconf_dict = load_synapse_configuration_data(syn_conf_path)

    return NrnMODPointProcessMechanismCustom(
        "synapse_mechs",
        synapses_data,
        synconf_dict,
        seed,
        rng_settings_mode,
        pre_mtypes,
        stim_params,
        use_glu_synapse=use_glu_synapse,
        syn_setup_params=syn_setup_params,
    )


def load_synapses_tsv_data(tsv_path):
    """Load synapse data from tsv.

    Args:
        tsv_path (str): path to the tsv synapses data file

    Returns:
        list of dicts containing each data for one synapse
    """
    synapses = []
    with open(tsv_path, "r", encoding="utf-8") as f:
        # first line is dimensions
        for line in f.readlines()[1:]:
            syn = {}
            items = line.strip().split("\t")
            syn["sid"] = int(items[0])
            syn["pre_cell_id"] = int(items[1])
            syn["sectionlist_id"] = int(items[2])
            syn["sectionlist_index"] = int(items[3])
            syn["seg_x"] = float(items[4])
            syn["synapse_type"] = int(items[5])
            syn["dep"] = float(items[6])
            syn["fac"] = float(items[7])
            syn["use"] = float(items[8])
            syn["tau_d"] = float(items[9])
            syn["delay"] = float(items[10])
            syn["weight"] = float(items[11])
            syn["Nrrp"] = float(items[12])
            syn["pre_mtype"] = int(items[13])

            synapses.append(syn)

    return synapses


def load_synapse_configuration_data(synconf_path):
    """Load synapse configuration data into dict[command]=list(ids).

    Args:
        synconf_path (str): path to the synapse configuration data file

    Returns:
        dict: configuration data

        each key contains a command to execute using hoc,
        and each value contains a list of synapse id on which to execute the command
    """
    synconf_dict = {}
    with open(synconf_path, "r", encoding="utf-8") as f:
        synconfs = f.read().split("-1000000000000000.0")

    for synconf in synconfs:
        tmp = synconf.split("\n")
        if "" in tmp:
            tmp.remove("")
        if len(tmp) == 2:
            cmd, ids = tmp
            ids = ids.replace(") ", ");")
            ids = ids.split(";")
            if "" in ids:
                ids.remove("")
            synconf_dict[cmd] = ids

    return synconf_dict



"""Configuration validation."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pprint
from pathlib import Path
from abc import ABC
from ast import literal_eval

from schema import Schema, And, Or

from emodelrunner.configuration.configparser import EModelConfigParser

logger = logging.getLogger(__name__)


class ConfigValidator(ABC):
    """Validates the config through a validation schema.

    Schema validation rules usage:
          - int: checks if the value is int.
          - Or("python", "neuron"): allows either "python" or "neuron".
          - And(str, len): string with length > 0
          - Any named function or lambda expression that evaluates to True is valid.
    """

    config_validator_schema = Schema({})
    default_values = {}

    @staticmethod
    def evaluates_to(n, data_type):
        """Checks if the expression n evaluates to the data_type.

        Args:
            n (str): the parameter value to be evaluated
            data_type (type): the datatype e.g. int, float, str, list

        Returns:
            bool: true n evaluates to the input data type, false otherwise.
        """
        return isinstance(literal_eval(n), data_type)

    @classmethod
    def int_expression(cls, n):
        """Check if n evaluates to an int literal.

        Args:
            n (str): the parameter value to be evaluated

        Returns:
            bool: true if the expression n evaluates to int, false otherwise
        """
        return cls.evaluates_to(n, int)

    @classmethod
    def float_or_int_expression(cls, n):
        """Check if n evaluates to a float or an integer literal.

        Args:
            n (str): the parameter value to be evaluated

        Returns:
            bool: true if the expression n evaluates to float, false otherwise
        """
        return cls.evaluates_to(n, float) or cls.evaluates_to(n, int)

    @staticmethod
    def list_of_nonempty_str(list_instance):
        """Check if the input is a list of nonempty strings.

        The list itself can be empty but it cannot contain an empty string.

        Args:
            list_instance (str): a string that evaluates to list.

        Returns:
            bool: true if the expression evaluates to list of non-empty strings.
        """
        list_instance = literal_eval(list_instance)
        return all(isinstance(s, str) and len(s) for s in list_instance)

    @staticmethod
    def boolean_expression(bool_input):
        """Checks if the expression has an expected boolean value.

        configparser.ConfigParser.getboolean() method evaluates this value.

        Args:
            bool_input (str): string containing boolean input.

        Returns:
            bool: True if the input is in the list of expected inputs.
        """
        return bool_input in ["True", "False", "true", "false", "1", "0"]

    def validate_from_file(self, config_path):
        """Validates the config at the given path and returns it.

        Args:
            config_path (str or Path): path to the .ini configuration file.

        Returns:
            configparser.ConfigParser: the validated config.

        Raises:
            FileNotFoundError: if config_path does not exist.
        """
        config = self._get_unvalidated_config(config_path)

        confdict = {
            section: dict(config.items(section)) for section in config.sections()
        }

        validated_conf = self.config_validator_schema.validate(confdict)
        logger.info("The loaded parameters are:")
        logger.info(pprint.pformat(validated_conf))

        logger.info("The config file is valid.")

        return config

    def _get_unvalidated_config(self, config_path):
        """Returns the config at the given path and fill unset values with default values.

        Args:
            config_path (str or Path): path to the .ini configuration file.

        Returns:
            configparser.ConfigParser: the config.

        Raises:
            FileNotFoundError: if config_path does not exist.
        """
        if not Path(config_path).exists():
            raise FileNotFoundError(f"config file at {config_path} is not found.")
        config = EModelConfigParser()

        # set defaults
        config.read_dict(self.default_values)

        config.read(config_path)
        return config


class SSCXConfigValidator(ConfigValidator):
    """Validates the SSCX config through a validation schema."""

    default_values = {
        "Package": {"type": "sscx"},
        "Cell": {
            "celsius": "34",
            "v_init": "-80",
            "gid": "0",
        },
        "Protocol": {
            # -1 means there is no apical point
            "apical_point_isec": "-1",
        },
        "Morphology": {
            "do_replace_axon": "True",
            # is only used for naming the output files
            "mtype": "",
        },
        "Sim": {
            "cvode_active": "False",
            "dt": "0.025",
        },
        "Synapses": {
            "add_synapses": "False",
            "seed": "846515",
            "rng_settings_mode": "Random123",  # can be "Random123" or "Compatibility"
            # name to use for the hoc synapse template
            "hoc_synapse_template_name": "hoc_synapses",
        },
        "Paths": {
            "memodel_dir": ".",
            "output_dir": "%(memodel_dir)s/python_recordings",
            "params_path": "%(memodel_dir)s/config/params/final.json",
            "units_path": "%(memodel_dir)s/config/features/units.json",
            "templates_dir": "%(memodel_dir)s/templates",
            "cell_template_path": "%(templates_dir)s/cell_template_neurodamus.jinja2",
            "run_hoc_template_path": "%(templates_dir)s/run_hoc.jinja2",
            "createsimulation_template_path": "%(templates_dir)s/createsimulation.jinja2",
            "synapses_template_path": "%(templates_dir)s/synapses.jinja2",
            "main_protocol_template_path": "%(templates_dir)s/main_protocol.jinja2",
            "features_hoc_template_path": "%(templates_dir)s/features.hoc",
            "replace_axon_hoc_path": "%(templates_dir)s/replace_axon_hoc.hoc",
            "syn_dir_for_hoc": "%(memodel_dir)s/synapses",
            "syn_dir": "%(memodel_dir)s/synapses",
            "syn_data_file": "synapses.tsv",
            "syn_conf_file": "synconf.txt",
            "syn_hoc_file": "synapses.hoc",
            "syn_mtype_map": "mtype_map.tsv",
            "simul_hoc_file": "createsimulation.hoc",
            "cell_hoc_file": "cell.hoc",
            "run_hoc_file": "run.hoc",
            "main_protocol_file": "main_protocol.hoc",
            "features_hoc_file": "features.hoc",
        },
    }

    def __init__(self):
        """Define the schema through validation rules."""
        self.config_validator_schema = Schema(
            {
                "Package": {"type": lambda n: n.lower() == "sscx"},
                "Cell": {
                    "celsius": self.float_or_int_expression,
                    "v_init": self.float_or_int_expression,
                    "gid": self.int_expression,
                    "emodel": And(str, len),
                },
                "Protocol": {
                    "apical_point_isec": self.int_expression,
                },
                "Morphology": {
                    "mtype": And(str, len),
                    "do_replace_axon": self.boolean_expression,
                },
                "Sim": {
                    "cvode_active": self.boolean_expression,
                    "dt": self.float_or_int_expression,
                },
                "Synapses": {
                    "add_synapses": self.boolean_expression,
                    "seed": self.int_expression,
                    "rng_settings_mode": Or("Random123", "Compatibility"),
                    "hoc_synapse_template_name": And(str, len),
                },
                "Paths": {
                    "morph_path": lambda n: Path(n).exists(),
                    "prot_path": lambda n: Path(n).exists(),
                    "features_path": lambda n: Path(n).exists(),
                    "unoptimized_params_path": lambda n: Path(n).exists(),
                    "memodel_dir": lambda n: Path(n).exists(),
                    "output_dir": lambda n: Path(n).exists(),
                    "params_path": lambda n: Path(n).exists(),
                    "units_path": lambda n: Path(n).exists(),
                    "templates_dir": lambda n: Path(n).exists(),
                    "cell_template_path": lambda n: Path(n).exists(),
                    "run_hoc_template_path": lambda n: Path(n).exists(),
                    "createsimulation_template_path": lambda n: Path(n).exists(),
                    "synapses_template_path": lambda n: Path(n).exists(),
                    "main_protocol_template_path": lambda n: Path(n).exists(),
                    "features_hoc_template_path": lambda n: Path(n).exists(),
                    "replace_axon_hoc_path": lambda n: Path(n).exists(),
                    "syn_dir_for_hoc": lambda n: Path(n).exists(),
                    "syn_dir": lambda n: Path(n).exists(),
                    "syn_data_file": And(str, len),
                    "syn_conf_file": And(str, len),
                    "syn_hoc_file": And(str, len),
                    "syn_mtype_map": And(str, len),
                    "simul_hoc_file": And(str, len),
                    "cell_hoc_file": And(str, len),
                    "run_hoc_file": And(str, len),
                    "main_protocol_file": And(str, len),
                    "features_hoc_file": And(str, len),
                },
            }
        )


class ThalamusConfigValidator(ConfigValidator):
    """Validates the Thalamus config through a validation schema."""

    default_values = {
        "Package": {"type": "thalamus"},
        "Cell": {
            "celsius": "34",
            "v_init": "-80",
            "gid": "0",
        },
        "Protocol": {
            # -1 means there is no apical point
            "apical_point_isec": "-1",
        },
        "Morphology": {
            "do_replace_axon": "True",
            # is only used for naming the output files
            "mtype": "",
        },
        "Sim": {
            "cvode_active": "False",
            "dt": "0.025",
        },
        "Synapses": {
            "add_synapses": "False",
            "seed": "846515",
            "rng_settings_mode": "Random123",  # can be "Random123" or "Compatibility"
        },
        "Paths": {
            "memodel_dir": ".",
            "output_dir": "%(memodel_dir)s/python_recordings",
            "params_path": "%(memodel_dir)s/config/params/final.json",
            "units_path": "%(memodel_dir)s/config/features/units.json",
            "syn_dir": "%(memodel_dir)s/synapses",
            "syn_data_file": "synapses.tsv",
            "syn_conf_file": "synconf.txt",
            "syn_mtype_map": "mtype_map.tsv",
        },
    }

    def __init__(self):
        """Define the schema through validation rules."""
        self.config_validator_schema = Schema(
            {
                "Package": {"type": lambda n: n.lower() == "thalamus"},
                "Cell": {
                    "celsius": self.float_or_int_expression,
                    "v_init": self.float_or_int_expression,
                    "gid": self.int_expression,
                    "emodel": And(str, len),
                },
                "Protocol": {
                    "apical_point_isec": self.int_expression,
                },
                "Morphology": {
                    "mtype": And(str, len),
                    "do_replace_axon": self.boolean_expression,
                },
                "Sim": {
                    "cvode_active": self.boolean_expression,
                    "dt": self.float_or_int_expression,
                },
                "Synapses": {
                    "add_synapses": self.boolean_expression,
                    "seed": self.int_expression,
                    "rng_settings_mode": Or("Random123", "Compatibility"),
                },
                "Paths": {
                    "morph_path": lambda n: Path(n).exists(),
                    "prot_path": lambda n: Path(n).exists(),
                    "features_path": lambda n: Path(n).exists(),
                    "unoptimized_params_path": lambda n: Path(n).exists(),
                    "memodel_dir": lambda n: Path(n).exists(),
                    "output_dir": lambda n: Path(n).exists(),
                    "params_path": lambda n: Path(n).exists(),
                    "units_path": lambda n: Path(n).exists(),
                    "syn_dir": lambda n: Path(n).exists(),
                    "syn_data_file": And(str, len),
                    "syn_conf_file": And(str, len),
                    "syn_mtype_map": And(str, len),
                },
            }
        )


class SynplasConfigValidator(ConfigValidator):
    """Validates the Synplas config through a validation schema."""

    default_values = {
        "Package": {"type": "synplas"},
        "Paths": {
            "memodel_dir": ".",
            "params_path": "%(memodel_dir)s/config/params/final.json",
            "synplas_fit_params_path": "%(memodel_dir)s/config/fit_params.json",
            "syn_dir": "%(memodel_dir)s/synapses",
            "syn_data_file": "synapses.tsv",
            "syn_conf_file": "synconf.txt",
            "synplas_output_path": "%(memodel_dir)s/output.h5",
            "pairsim_output_path": "%(memodel_dir)s/output.h5",
            "pairsim_precell_output_path": "%(memodel_dir)s/output_precell.h5",
            "syn_prop_path": "%(syn_dir)s/synapse_properties.json",
        },
        "Morphology": {
            "do_replace_axon": "True",
        },
        "Synapses": {
            "seed": "846515",
            "rng_settings_mode": "Random123",  # can be "Random123" or "Compatibility"
        },
    }

    def __init__(self):
        """Define the schema through validation rules."""
        self.config_validator_schema = Schema(
            {
                "Package": {"type": lambda n: n.lower() == "synplas"},
                "Cell": {
                    "celsius": self.float_or_int_expression,
                    "v_init": self.float_or_int_expression,
                    "emodel": And(str, len),
                    "precell_emodel": And(str, len),
                    "gid": self.int_expression,
                    "precell_gid": self.int_expression,
                },
                "Morphology": {
                    "do_replace_axon": self.boolean_expression,
                },
                "Paths": {
                    "morph_path": lambda n: Path(n).exists(),
                    "precell_morph_path": lambda n: Path(n).exists(),
                    "unoptimized_params_path": lambda n: Path(n).exists(),
                    "memodel_dir": lambda n: Path(n).exists(),
                    "params_path": lambda n: Path(n).exists(),
                    "precell_unoptimized_params_path": lambda n: Path(n).exists(),
                    "synplas_fit_params_path": lambda n: Path(n).exists(),
                    "syn_dir": lambda n: Path(n).exists(),
                    "syn_data_file": And(str, len),
                    "syn_conf_file": And(str, len),
                    "stimuli_path": lambda n: Path(n).exists(),
                    "spiketrain_path": lambda n: Path(n).exists(),
                    "syn_prop_path": lambda n: Path(n).exists(),
                    # cannot validate output paths before the files are created,
                    # so check that it is a str
                    "synplas_output_path": And(str, len),
                    "pairsim_output_path": And(str, len),
                    "pairsim_precell_output_path": And(str, len),
                },
                "Protocol": {
                    "tstop": self.float_or_int_expression,
                    "precell_amplitude": self.float_or_int_expression,
                    "precell_width": self.float_or_int_expression,
                    "precell_spikedelay": self.float_or_int_expression,
                },
                "Synapses": {
                    "seed": self.int_expression,
                    "rng_settings_mode": Or("Random123", "Compatibility"),
                },
                "SynapsePlasticity": {
                    "fastforward": self.float_or_int_expression,
                    "invivo": self.boolean_expression,
                    "base_seed": self.int_expression,
                    "synrec": self.list_of_nonempty_str,
                },
            }
        )


def get_validated_config(config_path):
    """Returns the validated config for the specified package type.

    Args:
        config_path (str or Path): path to the configuration file.

    Returns:
        configparser.ConfigParser: loaded config object
    """
    package_type = determine_package_type(config_path)

    if package_type == "sscx":
        conf_validator = SSCXConfigValidator()
    elif package_type == "thalamus":
        conf_validator = ThalamusConfigValidator()
    elif package_type == "synplas":
        conf_validator = SynplasConfigValidator()
    else:
        raise ValueError(f"Unsupported config type: {package_type}")

    validated_config = conf_validator.validate_from_file(config_path)
    return validated_config


def determine_package_type(config_path):
    """Returns the package type from the config file.

    Supports old synplas config files without the Package type.

    Args:
        config_path (str or Path): path to the configuration file.

    Returns:
        str: package type
    """
    # pylint: disable=protected-access
    unvalidated_config = ConfigValidator()._get_unvalidated_config(config_path)
    if "SynapsePlasticity" in unvalidated_config:
        return "synplas"
    else:
        return unvalidated_config.get("Package", "type").lower()



"""Class containing simulation for the GUI."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import numpy as np

from bluepyopt import ephys

from emodelrunner.recordings import RecordingCustom
from emodelrunner.cell import CellModelCustom
from emodelrunner.synapses.stimuli import NrnNetStimStimulusCustom
from emodelrunner.load import (
    load_config,
    load_syn_mechs,
    load_unoptimized_parameters,
    load_mechanisms,
    get_release_params,
)
from emodelrunner.morphology import create_morphology
from emodelrunner.synapses.create_locations import get_syn_locs


def section_coordinate_3d(sec, seg_pos):
    """Returns the 3d coordinates of a point in a section.

    Args:
        sec: neuron section
        seg_pos (float): postion of the segment os the section
            (should be between 0 and 1)

    Returns:
        list: 3d coordinates of a point in a section, or None if not found
    """
    n3d = sec.n3d()

    arc3d = [sec.arc3d(i) for i in range(n3d)]
    x3d = [sec.x3d(i) for i in range(n3d)]
    y3d = [sec.y3d(i) for i in range(n3d)]
    z3d = [sec.z3d(i) for i in range(n3d)]

    if not arc3d or seg_pos < 0 or seg_pos > 1:
        return None

    seg_pos = seg_pos * arc3d[-1]

    if seg_pos in arc3d:
        idx = arc3d.index(seg_pos)
        local_x = x3d[idx]
        local_y = y3d[idx]
        local_z = z3d[idx]
        return [local_x, local_y, local_z]
    else:
        for i, arc in enumerate(arc3d[1:]):
            # here, arc is arc3d[i + 1]
            if arc > seg_pos > arc3d[i] and arc - arc3d[i] != 0:
                proportion = (seg_pos - arc3d[i]) / (arc - arc3d[i])
                local_x = x3d[i] + proportion * (x3d[i + 1] - x3d[i])
                local_y = y3d[i] + proportion * (y3d[i + 1] - y3d[i])
                local_z = z3d[i] + proportion * (z3d[i + 1] - z3d[i])
                return [local_x, local_y, local_z]

    return None


def get_pos_and_color(sec, seg_pos, syn_type):
    """Returns the position and the synaptic type (0 for inhib. or 1 for excit.).

    Args:
        sec: neuron section
        seg_pos (float): postion of the segment os the section
            (should be between 0 and 1)
        syn_type (int): synaptic type. excitatory if >=100,
            inhibitory if <100

    Returns:
        list: first 3 numbers are the position, fourth is the synaptic type, None if pos not found
    """
    pos = section_coordinate_3d(sec, seg_pos)

    if pos is None:
        return None

    if syn_type < 100:
        syn_type_ = 0
    else:
        syn_type_ = 1
    pos.append(syn_type_)

    return pos


def get_step_data(steps, step, default_step):
    """Extract step data from StepProtocol json dict and add amplitude to a step list.

    Args:
        steps (list): list of step amplitudes (nA) to be updated
        step (dict or list of dicts): step from which to extract step data
        default_step (float): default value for the custom step entry (nA)

    Returns:
        a tuple containing

        - float: total duration of the step (ms)
        - float: delay of the step (ms)
        - float: duration of the step (ms)
    """
    # can be dict or list of dicts
    if isinstance(step, list):
        for step_ in step:
            # If a default step amplitude is registered,
            # two buttons will have the same value:
            # the registered one and the custom one
            if step_["amp"] != default_step:
                steps.append(step_["amp"])

            # replace default values
            # may be replaced several times
            total_duration = step_["totduration"]
            step_delay = step_["delay"]
            step_duration = step_["duration"]
    else:
        if step["amp"] != default_step:
            steps.append(step["amp"])

        total_duration = step["totduration"]
        step_delay = step["delay"]
        step_duration = step["duration"]

    return total_duration, step_delay, step_duration


def get_holding_data(holdings, stimulus_data, total_duration, default_holding):
    """Extract holding data from StepProtocol json dict and add amplitude to a holding list.

    Args:
        holdings (list): list of holding amplitudes (nA) to be updated
        stimulus_data (dict): stimulus dict from protocol json file containing holding data
        total_duration (float): total duration of the step (ms)
        default_holding (float): default value for the custom holding entry

    Returns:
        a tuple containing

        - float: delay of the holding stimulus (ms)
        - float: duration of the holding stimulus (ms)
    """
    if "holding" in stimulus_data:
        holding = stimulus_data["holding"]

        # amp can be None in e.g. Rin recipe protocol
        if holding["amp"] is not None and holding["amp"] != default_holding:
            holdings.append(holding["amp"])

        hold_step_delay = holding["delay"]
        hold_step_duration = holding["duration"]
    else:
        hold_step_delay = 0.0
        hold_step_duration = total_duration

    return hold_step_delay, hold_step_duration


class NeuronSimulation:
    """Class containing BPO cell, simulation & protocol.

    Attributes:
        config (dict): dictionary containing configuration data
        cell_path (str): path to cell repo. should be "."
        total_duration (float): duration of cell simulation (ms)
        steps (list of floats): default step stimuli (nA)
        hypamps (list of floats): default holding stimuli (nA)
        step_stim (float): selected step stimulus (nA)
        hypamp (float): selected holding stimulus (nA)
        step_delay (float): delay before applying step stimulus (ms)
        step_duration (float): duration of step stimulus (ms)
        hold_step_delay (float): delay of holding stimulus (ms)
        hold_step_duration (float): duration of holding stimulus (ms)
        available_pre_mtypes (dict): all synapses pre_mtypes
            {mtypeidx: mtype_name, ...}
        pre_mtypes (list of int): selected pre_mtypes to run
            [mtypeidx, ...]
        netstim_params (dict): netstim parameters for synapses of each mtype
            {mtypeidx:[start, interval, number, noise]}
        syn_start (int): default time (ms) at which the synapse starts firing
        syn_interval (int): default interval (ms) between two synapse firing
        syn_nmb_of_spikes (int): default number of synapse firing
        syn_noise (int): default synapse noise
        protocol (ephys.protocols.SweepProtocol): BluePyOpt-based Protocol
        cell (CellModelCustom): BluePyOpt-based cell
        release_params (dict): optimised cell parameters to fill in
            the cell's free parameters
        sim (ephys.simulators.NrnSimulator): BluePyOpt simulator
            can access neuron data from it
        syn_display_data (dict): synapse data (position and type) for display
            syn_display_data[pre_mtype] = [x,y,z,type],
            type=0 if inhib, type=1 if excit
    """

    def __init__(self, config_path="config/config_allsteps.ini"):
        """Constructor. Load default params from config file.

        Args:
            config_path (str):path to the config file
        """
        # load config file
        self.config = load_config(config_path=config_path)
        self.cell_path = self.config.get("Paths", "memodel_dir")

        # get default params
        self.load_protocol_params()
        self.load_synapse_params()

        # uninstantiated params
        self.protocol = None
        self.cell = None
        self.release_params = None
        self.sim = None
        self.syn_display_data = None

    def load_protocol_params(
        self,
        total_duration=3000,
        step_delay=700,
        step_duration=2000,
        hold_step_delay=0,
        hold_step_duration=3000,
        default_step=0,
        default_holding=0,
    ):
        """Load default protocol params.

        Args:
            total_duration (float): default value for duration of cell simulation (ms)
                if no StepProtocol is found
            step_delay (float): default value for delay before applying step stimulus (ms)
                if no StepProtocol is found
            step_duration (float): default value for duration of step stimulus (ms)
                if no StepProtocol is found
            hold_step_delay (float): default value for delay of holding stimulus (ms)
                if no StepProtocol is found
            hold_step_duration (float): default value for duration of holding stimulus (ms)
                if no StepProtocol is found
            default_step (float): default value for custom step amplitude (nA)
            default_holding (float): default value for custom holding amplitude (nA)
        """
        prot_path = self.config.get("Paths", "prot_path")
        with open(prot_path, "r", encoding="utf-8") as protocol_file:
            protocol_data = json.load(protocol_file)
        if "__comment" in protocol_data:
            del protocol_data["__comment"]

        # list of all steps and hold amps found in all stepprotocols in prot file
        steps = []
        holdings = []

        for prot_data in protocol_data.values():
            # update default delays / durations and update steps and holdings lists
            if prot_data["type"] == "StepProtocol":
                total_duration, step_delay, step_duration = get_step_data(
                    steps=steps,
                    step=prot_data["stimuli"]["step"],
                    default_step=default_step,
                )

                hold_step_delay, hold_step_duration = get_holding_data(
                    holdings, prot_data["stimuli"], total_duration, default_holding
                )

        self.total_duration = total_duration

        # filter duplicates (dict preserves order for py37+)
        self.steps = list(dict.fromkeys(steps))
        self.hypamps = list(dict.fromkeys(holdings))

        # set default values for custom entry
        self.step_stim = default_step
        self.hypamp = default_holding

        self.step_delay = step_delay
        self.step_duration = step_duration
        self.hold_step_delay = hold_step_delay
        self.hold_step_duration = hold_step_duration

    def load_synapse_params(
        self, syn_start=0, syn_interval=0, syn_nmb_of_spikes=0, syn_noise=0
    ):
        """Load default synapse params.

        Args:
            syn_start (int): default time (ms) at which the synapse starts firing
            syn_interval (int): default interval (ms) between two synapse firing
            syn_nmb_of_spikes (int): default number of synapse firing
            syn_noise (int): default synapse noise
        """
        # mtypes to be chosen from {mtypeidx: mtype_name, ...}
        self.available_pre_mtypes = self.load_available_pre_mtypes()
        # mtypes to be loaded [mtypeidx, ...]
        self.pre_mtypes = []
        # synapse netstim param depending on mtype {mtypeidx:[start, interval, number, noise]}
        self.netstim_params = {}

        # default synapse params
        self.syn_start = syn_start
        self.syn_interval = syn_interval
        self.syn_nmb_of_spikes = syn_nmb_of_spikes
        self.syn_noise = syn_noise

    def load_available_pre_mtypes(self):
        """Load the list of pre mtype cells to which are connected the synapses.

        Returns:
            dict: mtypes of cells connected to the synapses
        """
        mtype_path = os.path.join(
            self.config.get("Paths", "syn_dir"),
            self.config.get("Paths", "syn_mtype_map"),
        )
        with open(mtype_path, "r", encoding="utf-8") as mtype_file:
            raw_mtypes = mtype_file.readlines()

        # mtypes[id] = m-type name
        mtypes = {}
        for line in raw_mtypes:
            line = line.rstrip().split()
            if line:
                mtypes[int(line[0])] = line[1]

        return mtypes

    def get_syn_stim(self):
        """Create synapse stimuli.

        Returns:
            emodelrunner.synapses.stimuli.NrnNetStimStimulusCustom: synapse stimuli
        """
        if self.pre_mtypes:
            syn_locs = get_syn_locs(self.cell)
            syn_total_duration = self.total_duration
            return NrnNetStimStimulusCustom(syn_locs, syn_total_duration)
        return None

    def load_protocol(self, protocol_name="protocol"):
        """Load BPO protocol.

        Args:
            protocol_name (str): protocol name to use in BluePyOpt classes.
                Does not have an effect on the simulation
        """
        syn_stim = self.get_syn_stim()

        soma_loc = ephys.locations.NrnSeclistCompLocation(
            name="soma", seclist_name="somatic", sec_index=0, comp_x=0.5
        )

        rec = RecordingCustom(name=protocol_name, location=soma_loc, variable="v")

        # create step stimulus
        stim = ephys.stimuli.NrnSquarePulse(
            step_amplitude=self.step_stim,
            step_delay=self.step_delay,
            step_duration=self.step_duration,
            location=soma_loc,
            total_duration=self.total_duration,
        )

        # create holding stimulus
        hold_stim = ephys.stimuli.NrnSquarePulse(
            step_amplitude=self.hypamp,
            step_delay=self.hold_step_delay,
            step_duration=self.hold_step_duration,
            location=soma_loc,
            total_duration=self.total_duration,
        )

        # create protocol
        stims = [stim, hold_stim]
        if syn_stim is not None:
            stims.append(syn_stim)

        self.protocol = ephys.protocols.SweepProtocol(
            protocol_name, stims, [rec], False
        )

    def create_cell_custom(self):
        """Create a cell.

        Returns:
            emodelrunner.cell.CellModelCustom: cell model
        """
        # pylint: disable=too-many-locals
        emodel = self.config.get("Cell", "emodel")
        gid = self.config.getint("Cell", "gid")

        # load mechanisms
        unopt_params_path = self.config.get("Paths", "unoptimized_params_path")
        mechs = load_mechanisms(unopt_params_path)

        # add synapses mechs
        seed = self.config.getint("Synapses", "seed")
        rng_settings_mode = self.config.get("Synapses", "rng_settings_mode")
        syn_data_path = os.path.join(
            self.config.get("Paths", "syn_dir"),
            self.config.get("Paths", "syn_data_file"),
        )
        syn_conf_path = os.path.join(
            self.config.get("Paths", "syn_dir"),
            self.config.get("Paths", "syn_conf_file"),
        )
        # always load synapse data for synapse display.
        # -> do not need to reload syn data each time user toggles synapse checkbox
        mechs += [
            load_syn_mechs(
                seed,
                rng_settings_mode,
                syn_data_path,
                syn_conf_path,
                self.pre_mtypes,
                self.netstim_params,
            )
        ]

        # load parameters
        params = load_unoptimized_parameters(
            unopt_params_path,
            v_init=self.config.getfloat("Cell", "v_init"),
            celsius=self.config.getfloat("Cell", "celsius"),
        )

        # load morphology
        morph_config = self.config.morph_args()
        morph = create_morphology(morph_config, self.config.package_type)

        # create cell
        cell = CellModelCustom(
            name=emodel,
            morph=morph,
            mechs=mechs,
            params=params,
            gid=gid,
        )

        return cell

    def load_cell_sim(self):
        """Load BPO cell & simulation."""
        self.cell = self.create_cell_custom()
        self.release_params = get_release_params(self.config)
        self.sim = ephys.simulators.NrnSimulator(
            dt=self.config.getfloat("Sim", "dt"), cvode_active=False
        )

    def load_synapse_display_data(self):
        """Load dict containing x,y,z of each synapse & inhib/excit."""
        # self.syn_display_data[pre_mtype] = [x,y,z,type], type=0 if inhib, type=1 if excit
        self.syn_display_data = {}
        for key in self.available_pre_mtypes:
            self.syn_display_data[key] = []

        for mech in self.cell.mechanisms:
            if hasattr(mech, "pprocesses"):
                for syn in mech.synapses_data:
                    pre_mtype = syn["pre_mtype"]
                    seg_pos = syn["seg_x"]
                    # check if a synapse of the same mtype has already the same position
                    # and add synapse only if a new position has to be displayed
                    syn_section = mech.get_cell_section_for_synapse(
                        syn, self.cell.icell
                    )
                    syn_display_data = get_pos_and_color(
                        syn_section, seg_pos, syn["synapse_type"]
                    )
                    if (
                        syn_display_data is not None
                        and syn_display_data not in self.syn_display_data[pre_mtype]
                    ):
                        self.syn_display_data[pre_mtype].append(syn_display_data)

    def instantiate(self):
        """Instantiate cell, simulation & protocol."""
        self.cell.freeze(self.release_params)
        self.cell.instantiate(sim=self.sim)
        self.protocol.instantiate(sim=self.sim, cell_model=self.cell)
        self.sim.neuron.h.tstop = self.protocol.total_duration
        self.sim.neuron.h.stdinit()

    def destroy(self):
        """Destroy cell & protocol."""
        self.protocol.destroy(sim=self.sim)
        self.cell.destroy(sim=self.sim)
        self.cell.unfreeze(self.release_params.keys())

    def get_voltage(self):
        """Returns voltage response.

        Returns:
            a tuple containing

            - ndarray: the time response
            - ndarray: the voltage response
        """
        responses = {
            recording.name: recording.response for recording in self.protocol.recordings
        }
        key = list(responses.keys())[0]
        resp = responses[key]
        return np.array(resp["time"]), np.array(resp["voltage"])



"""Protocol creation functions & custom protocol classes for Synplas packages."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import logging
import sys
import traceback

from bluepyopt import ephys

logger = logging.getLogger(__name__)


def fastforward_synapses(cell_model):
    """Enable synapse fast-forwarding.

    Args:
        cell_model (emodelrunner.cell.CellModelCustom): the cell whose synapses
            are to be fast-forwarded
    """
    for mech in cell_model.mechanisms:
        if hasattr(mech, "pprocesses"):
            for synapse in mech.pprocesses:
                if synapse.hsynapse.rho_GB >= 0.5:
                    synapse.hsynapse.rho_GB = 1.0
                    synapse.hsynapse.Use_TM = synapse.hsynapse.Use_p_TM
                    synapse.hsynapse.gmax_AMPA = synapse.hsynapse.gmax_p_AMPA
                else:
                    synapse.hsynapse.rho_GB = 0.0
                    synapse.hsynapse.Use_TM = synapse.hsynapse.Use_d_TM
                    synapse.hsynapse.gmax_AMPA = synapse.hsynapse.gmax_d_AMPA


class SweepProtocolCustom(ephys.protocols.SweepProtocol):
    """SweepProtocol with ability of synapse fastforwarding.

    Attributes:
        name (str): name of this object.
        stimuli (list of 2 lists of Stimuli): Stimulus objects used in the protocol
            The list must be of size 2 and
            contain first the list for the presynaptic stimuli
            and then the list for the postsynaptic stimuli.
        recordings (list of 2 lists of Recordings): Recording objects used in the
            protocol. The list must be of size 2 and
            contain first the list for the presynaptic recording
            and then the list for the postsynaptic recording.
        cvode_active (bool): whether to use variable time step
        fastforward (float): Time after which the synapses are fasforwarded.
            Leave None for no fastforward.
    """

    def __init__(
        self,
        name=None,
        stimuli=None,
        recordings=None,
        cvode_active=None,
        fastforward=None,
    ):
        """Constructor.

        Args:
            name (str): name of this object.
            stimuli (list of 2 lists of Stimuli): Stimulus objects used in the protocol
                The list must be of size 2 and
                contain first the list for the presynaptic stimuli
                and then the list for the postsynaptic stimuli.
            recordings (list of 2 lists of Recordings): Recording objects used in the
                protocol. The list must be of size 2 and
                contain first the list for the presynaptic recording
                and then the list for the postsynaptic recording.
            cvode_active (bool): whether to use variable time step
            fastforward (float): Time after which the synapses are fasforwarded.
                Leave None for no fastforward.
        """
        super().__init__(name, stimuli, recordings, cvode_active)

        self.fastforward = fastforward

    def _run_func(self, cell_model, param_values, sim=None):
        """Run protocols.

        Args:
            cell_model (bluepyopt.ephys.models.CellModel): the cell model
            param_values (dict): optimized parameters
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator

        Raises:
            Exception: if the instantiation failed

        Returns:
            dict containing the responses
        """
        # pylint: disable=raise-missing-from
        try:
            cell_model.freeze(param_values)
            cell_model.instantiate(sim=sim)

            self.instantiate(sim=sim, cell_model=cell_model)

            try:
                if self.fastforward is not None:
                    sim.run(self.fastforward, cvode_active=self.cvode_active)
                    fastforward_synapses(cell_model)
                    sim.neuron.h.cvode_active(1)
                    sim.neuron.h.continuerun(self.total_duration)
                else:
                    sim.run(self.total_duration, cvode_active=self.cvode_active)
            except (RuntimeError, ephys.simulators.NrnSimulatorException):
                logger.debug(
                    "SweepProtocol: Running of parameter set {%s} generated "
                    "an exception, returning None in responses",
                    str(param_values),
                )
                responses = {recording.name: None for recording in self.recordings}
            else:
                responses = {
                    recording.name: recording.response for recording in self.recordings
                }

            self.destroy(sim=sim)

            cell_model.destroy(sim=sim)

            cell_model.unfreeze(param_values.keys())

            return responses
        except BaseException:
            raise Exception("".join(traceback.format_exception(*sys.exc_info())))


class SweepProtocolPairSim(ephys.protocols.Protocol):
    """Sweep protocol for pair simulation with fastforwarding.

    Attributes:
        name (str): name of this object.
        stimuli (list of 2 lists of Stimuli): Stimulus objects used in the protocol
            The list must be of size 2 and
            contain first the list for the presynaptic stimuli
            and then the list for the postsynaptic stimuli.
        recordings (list of 2 lists of Recordings): Recording objects used in the
            protocol. The list must be of size 2 and
            contain first the list for the presynaptic recording
            and then the list for the postsynaptic recording.
        cvode_active (bool): whether to use variable time step
        fastforward (float): Time after which the synapses are fasforwarded.
            Leave None for no fastforward.
    """

    def __init__(
        self,
        name=None,
        stimuli=None,
        recordings=None,
        cvode_active=None,
        fastforward=None,
    ):
        """Constructor.

        Args:
            name (str): name of this object.
            stimuli (list of 2 lists of Stimuli): Stimulus objects used in the protocol
                The list must be of size 2 and
                contain first the list for the presynaptic stimuli
                and then the list for the postsynaptic stimuli.
            recordings (list of 2 lists of Recordings): Recording objects used in the
                protocol. The list must be of size 2 and
                contain first the list for the presynaptic recording
                and then the list for the postsynaptic recording.
            cvode_active (bool): whether to use variable time step
            fastforward (float): Time after which the synapses are fasforwarded.
                Leave None for no fastforward.

        Raises:
            Exception: if stimuli is not of size 2 and is not None
            Exception: if recordings is not of size 2 and is not None
        """
        super().__init__(name)
        if stimuli is not None and len(stimuli) != 2:
            raise Exception(
                "Stimuli should be of size 2 and contain"
                "[presynaptic_stimuli, postsynaptic_stimuli]"
            )
        self.stimuli = stimuli
        if recordings is not None and len(stimuli) != 2:
            raise Exception(
                "Recordings should be of size 2 and contain"
                "[presynaptic_recordings, postsynaptic_recordings]"
            )
        self.recordings = recordings
        self.cvode_active = cvode_active
        self.fastforward = fastforward

    @property
    def total_duration(self):
        """Total duration.

        Returns:
            float: total duration
        """
        return max(
            stimulus.total_duration
            for stimulus_sublist in self.stimuli
            for stimulus in stimulus_sublist
        )

    def subprotocols(self):
        """Return subprotocols.

        Returns:
            a dict containing the object
        """
        return collections.OrderedDict({self.name: self})

    def _run_func(
        self,
        precell_model,
        postcell_model,
        pre_param_values,
        post_param_values,
        sim=None,
    ):
        """Run protocols.

        Args:
            precell_model (bluepyopt.ephys.models.CellModel): the presynaptic cell model
            postcell_model (bluepyopt.ephys.models.CellModel): the postsynaptic cell model
            pre_param_values (dict): optimized parameters of the presynaptic cell model
            post_param_values (dict): optimized parameters of the postsynaptic cell model
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator

        Raises:
            Exception: if the instantiation failed

        Returns:
            list of 2 dicts containing the responses of both the cells

            Has the structure [presynaptic response dict, postsynaptic response dict]
        """
        # pylint: disable=raise-missing-from
        try:
            precell_model.freeze(pre_param_values)
            precell_model.instantiate(sim=sim)

            postcell_model.freeze(post_param_values)
            postcell_model.instantiate(sim=sim)

            self.instantiate(
                sim=sim, pre_icell=precell_model.icell, post_icell=postcell_model.icell
            )

            try:
                if self.fastforward is not None:
                    sim.run(self.fastforward, cvode_active=self.cvode_active)
                    fastforward_synapses(precell_model)
                    fastforward_synapses(postcell_model)
                    sim.neuron.h.cvode_active(1)
                    sim.neuron.h.continuerun(self.total_duration)
                else:
                    sim.run(self.total_duration, cvode_active=self.cvode_active)
            except (RuntimeError, ephys.simulators.NrnSimulatorException):
                logger.debug(
                    "SweepProtocol: Running of parameter sets {%s} and {%s} generated "
                    "an exception, returning None in responses",
                    str(pre_param_values),
                    str(post_param_values),
                )
                responses = [
                    {recording.name: None for recording in recordings}
                    for recordings in self.recordings
                ]
            else:
                responses = [
                    {recording.name: recording.response for recording in recordings}
                    for recordings in self.recordings
                ]

            self.destroy(sim=sim)

            precell_model.destroy(sim=sim)
            postcell_model.destroy(sim=sim)

            precell_model.unfreeze(pre_param_values.keys())
            postcell_model.unfreeze(post_param_values.keys())

            return responses
        except BaseException:
            raise Exception("".join(traceback.format_exception(*sys.exc_info())))

    def run(
        self,
        precell_model,
        postcell_model,
        pre_param_values,
        post_param_values,
        sim=None,
        isolate=None,
        timeout=None,
    ):
        """Instantiate protocol.

        Args:
            precell_model (bluepyopt.ephys.models.CellModel): the presynaptic cell model
            postcell_model (bluepyopt.ephys.models.CellModel): the postsynaptic cell model
            pre_param_values (dict): optimized parameters of the presynaptic cell model
            post_param_values (dict): optimized parameters of the postsynaptic cell model
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            isolate (bool): whether to isolate the run in a process with a timeout
                to avoid bad cells running for too long
            timeout (float): maximum real time (s) the cells are allowed to run when isolated

        Returns:
            list of 2 dicts containing the responses of both the cells

            Has the structure [presynaptic response dict, postsynaptic response dict]
        """
        # pylint:disable=too-many-locals, import-outside-toplevel
        if isolate is None:
            isolate = True

        if isolate:

            def _reduce_method(meth):
                """Overwrite reduce."""
                return (getattr, (meth.__self__, meth.__func__.__name__))

            import copyreg
            import types

            copyreg.pickle(types.MethodType, _reduce_method)
            import pebble
            from concurrent.futures import TimeoutError as FuturesTimeoutError

            if timeout is not None:
                if timeout < 0:
                    raise ValueError("timeout should be > 0")

            with pebble.ProcessPool(max_workers=1, max_tasks=1) as pool:
                tasks = pool.schedule(
                    self._run_func,
                    kwargs={
                        "precell_model": precell_model,
                        "postcell_model": postcell_model,
                        "pre_param_values": pre_param_values,
                        "post_param_values": post_param_values,
                        "sim": sim,
                    },
                    timeout=timeout,
                )
                try:
                    responses = tasks.result()
                except FuturesTimeoutError:
                    logger.debug(
                        "SweepProtocol: task took longer than "
                        "timeout, will return empty response "
                        "for this recording"
                    )
                    responses = [
                        {recording.name: None for recording in recordings}
                        for recordings in self.recordings
                    ]
        else:
            responses = self._run_func(
                precell_model=precell_model,
                postcell_model=postcell_model,
                pre_param_values=pre_param_values,
                post_param_values=post_param_values,
                sim=sim,
            )
        return responses

    def instantiate(self, sim=None, pre_icell=None, post_icell=None):
        """Instantiate.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
            pre_icell (neuron cell): presynaptic cell instantiation in simulator
            post_icell (neuron cell): postsynaptic cell instantiation in simulator
        """
        icells = [pre_icell, post_icell]

        for i, icell in enumerate(icells):

            for stimulus in self.stimuli[i]:
                stimulus.instantiate(sim=sim, icell=icell)

            for recording in self.recordings[i]:
                try:
                    recording.instantiate(sim=sim, icell=icell)
                except ephys.locations.EPhysLocInstantiateException:
                    logger.debug(
                        "SweepProtocol: Instantiating recording generated "
                        "location exception, will return empty response for "
                        "this recording"
                    )

    def destroy(self, sim=None):
        """Destroy protocol.

        Args:
            sim (bluepyopt.ephys.NrnSimulator): neuron simulator
        """
        for stimulus_list in self.stimuli:
            for stimulus in stimulus_list:
                stimulus.destroy(sim=sim)

        for recording_list in self.recordings:
            for recording in recording_list:
                recording.destroy(sim=sim)

    def __str__(self):
        """String representation.

        Returns:
            str describing the stimuli and recordings
        """
        content = f"{self.name}:\n"

        content += "  pre-synaptic stimuli:\n"
        for stimulus in self.stimuli[0]:
            content += f"    {stimulus}\n"

        content += "  post-synaptic stimuli:\n"
        for stimulus in self.stimuli[1]:
            content += f"    {stimulus}\n"

        content += "  pre-synaptic recordings:\n"
        for recording in self.recordings[0]:
            content += f"    {recording}\n"

        content += "  post-synaptic recordings:\n"
        for recording in self.recordings[1]:
            content += f"    {recording}\n"

        return content



#!/usr/bin/env python
"""EModelRunner setup."""

# Copyright 2020-2022 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 7):
    sys.exit("Sorry, Python < 3.7 is not supported")

# read the contents of the README file
with open("README.rst", encoding="utf-8") as f:
    README = f.read()

setup(
    name="EModelRunner",
    author="Blue Brain Project, EPFL",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Runs cells from Blue Brain Project cell packages, such as sscx, synapse plasticity, etc.",
    long_description=README,
    long_description_content_type="text/x-rst",
    url="https://github.com/BlueBrain/EModelRunner",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/EModelRunner/issues",
        "Source": "https://github.com/BlueBrain/EModelRunner",
        "Documentation": "https://emodelrunner.readthedocs.io/en/latest",
    },
    license="Apache 2.0",
    install_requires=[
        "numpy",
        "bluepyopt>=1.13.168",
        "neurom>=3.1.0",
        "h5py",
        "matplotlib",
        "schema",
        "Pebble>=4.3.10",
        "importlib_metadata; python_version<'3.8'",
    ],
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.7",
    extras_require={"docs": ["sphinx", "sphinx-bluebrain-theme"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    include_package_data=True,
)
